@@ -25,6 +25,7 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/client/compositor.h"
 #include "../../src/client/connection_thread.h"
 #include "../../src/client/event_queue.h"
+#include "../../src/client/output.h"
 #include "../../src/client/surface.h"
 #include "../../src/client/region.h"
 #include "../../src/client/registry.h"
@@ -59,6 +60,7 @@ private Q_SLOTS:
     void testSurfaceAt();
     void testDestroyAttachedBuffer();
     void testDestroyWithPendingCallback();
+    void testOutput();
     void testDisconnect();
 
 private:
@@ -259,7 +261,7 @@ void TestWaylandSurface::testDamage()
     QVERIFY(damageSpy.isEmpty());
     QVERIFY(!serverSurface->isMapped());
 
-    QImage img(QSize(10, 10), QImage::Format_ARGB32);
+    QImage img(QSize(10, 10), QImage::Format_ARGB32_Premultiplied);
     img.fill(Qt::black);
     auto b = m_shm->createBuffer(img);
     s->attachBuffer(b);
@@ -273,7 +275,7 @@ void TestWaylandSurface::testDamage()
     // damage multiple times
     QRegion testRegion(5, 8, 3, 6);
     testRegion = testRegion.united(QRect(10, 20, 30, 15));
-    img = QImage(QSize(40, 35), QImage::Format_ARGB32);
+    img = QImage(QSize(40, 35), QImage::Format_ARGB32_Premultiplied);
     img.fill(Qt::black);
     b = m_shm->createBuffer(img);
     s->attachBuffer(b);
@@ -300,7 +302,7 @@ void TestWaylandSurface::testFrameCallback()
 
     QSignalSpy frameRenderedSpy(s, SIGNAL(frameRendered()));
     QVERIFY(frameRenderedSpy.isValid());
-    QImage img(QSize(10, 10), QImage::Format_ARGB32);
+    QImage img(QSize(10, 10), QImage::Format_ARGB32_Premultiplied);
     img.fill(Qt::black);
     auto b = m_shm->createBuffer(img);
     s->attachBuffer(b);
@@ -323,10 +325,10 @@ void TestWaylandSurface::testAttachBuffer()
     KWayland::Server::SurfaceInterface *serverSurface = serverSurfaceCreated.first().first().value<KWayland::Server::SurfaceInterface*>();
     QVERIFY(serverSurface);
 
-    // create two images
+    // create three images
     QImage black(24, 24, QImage::Format_RGB32);
     black.fill(Qt::black);
-    QImage red(24, 24, QImage::Format_ARGB32);
+    QImage red(24, 24, QImage::Format_ARGB32); //Note - deliberately not premultiplied
     red.fill(QColor(255, 0, 0, 128));
     QImage blue(24, 24, QImage::Format_ARGB32_Premultiplied);
     blue.fill(QColor(0, 0, 255, 128));
@@ -369,8 +371,15 @@ void TestWaylandSurface::testAttachBuffer()
     KWayland::Server::BufferInterface *buffer2 = serverSurface->buffer();
     buffer2->ref();
     QVERIFY(buffer2->shmBuffer());
-    QCOMPARE(buffer2->data(), red);
-    QCOMPARE(buffer2->data().format(), QImage::Format_ARGB32);
+    QCOMPARE(buffer2->data().format(), QImage::Format_ARGB32_Premultiplied);
+    QCOMPARE(buffer2->data().width(), 24);
+    QCOMPARE(buffer2->data().height(), 24);
+    for (int i = 0; i < 24; ++i) {
+        for (int j = 0; j < 24; ++j) {
+            // it's premultiplied in the format
+            QCOMPARE(buffer2->data().pixel(i, j), qRgba(128, 0, 0, 128));
+        }
+    }
     buffer2->unref();
     QVERIFY(buffer2->isReferenced());
     QVERIFY(!redBuffer.data()->isReleased());
@@ -398,7 +407,7 @@ void TestWaylandSurface::testAttachBuffer()
     KWayland::Server::BufferInterface *buffer3 = serverSurface->buffer();
     buffer3->ref();
     QVERIFY(buffer3->shmBuffer());
-    QCOMPARE(buffer3->data().format(), QImage::Format_ARGB32);
+    QCOMPARE(buffer3->data().format(), QImage::Format_ARGB32_Premultiplied);
     QCOMPARE(buffer3->data().width(), 24);
     QCOMPARE(buffer3->data().height(), 24);
     for (int i = 0; i < 24; ++i) {
@@ -480,7 +489,7 @@ void TestWaylandSurface::testMultipleSurfaces()
     // create two images
     QImage black(24, 24, QImage::Format_RGB32);
     black.fill(Qt::black);
-    QImage red(24, 24, QImage::Format_ARGB32);
+    QImage red(24, 24, QImage::Format_ARGB32_Premultiplied);
     red.fill(QColor(255, 0, 0, 128));
 
     auto blackBuffer = pool1.createBuffer(black);
@@ -669,6 +678,10 @@ void TestWaylandSurface::testScale()
 
     // let's change the scale factor
     QSignalSpy scaleChangedSpy(serverSurface, &SurfaceInterface::scaleChanged);
+
+    //changing the scale implicitly changes the size
+    QSignalSpy sizeChangedSpy(serverSurface, &SurfaceInterface::sizeChanged);
+
     QVERIFY(scaleChangedSpy.isValid());
     s->setScale(2);
     QCOMPARE(s->scale(), 2);
@@ -680,6 +693,11 @@ void TestWaylandSurface::testScale()
     QCOMPARE(scaleChangedSpy.first().first().toInt(), 2);
     QCOMPARE(serverSurface->scale(), 2);
 
+    //even though we've changed the scale, if we don't have a buffer we
+    //don't have a size. If we don't have a size it can't have changed
+    QCOMPARE(sizeChangedSpy.count(), 0);
+    QVERIFY(!serverSurface->size().isValid());
+
     // let's try changing to same factor, should not emit changed on server
     s->setScale(2);
     s->commit(Surface::CommitFlag::None);
@@ -693,6 +711,44 @@ void TestWaylandSurface::testScale()
     QCOMPARE(scaleChangedSpy.first().first().toInt(), 2);
     QCOMPARE(scaleChangedSpy.last().first().toInt(), 4);
     QCOMPARE(serverSurface->scale(), 4);
+    scaleChangedSpy.clear();
+
+    //attach a buffer of 100x100, our scale is 4, so this should be a size of 25x25
+    QImage red(100, 100, QImage::Format_ARGB32_Premultiplied);
+    red.fill(QColor(255, 0, 0, 128));
+    auto redBuffer = m_shm->createBuffer(red);
+    s->attachBuffer(redBuffer.data());
+    s->damage(QRect(0,0, 25,25));
+    s->commit(Surface::CommitFlag::None);
+    QVERIFY(sizeChangedSpy.wait());
+    QCOMPARE(sizeChangedSpy.count(), 1);
+    QCOMPARE(serverSurface->size(), QSize(25,25));
+    sizeChangedSpy.clear();
+    scaleChangedSpy.clear();
+
+    //set the scale to 1, buffer is still 100x100 so size should change to 100x100
+    s->setScale(1);
+    s->commit(Surface::CommitFlag::None);
+    QVERIFY(sizeChangedSpy.wait());
+    QCOMPARE(sizeChangedSpy.count(), 1);
+    QCOMPARE(scaleChangedSpy.count(), 1);
+    QCOMPARE(serverSurface->scale(), 1);
+    QCOMPARE(serverSurface->size(), QSize(100,100));
+    sizeChangedSpy.clear();
+    scaleChangedSpy.clear();
+
+    //set scale and size in one commit, buffer is 50x50 at scale 2 so size should be 25x25
+    QImage blue(50, 50, QImage::Format_ARGB32_Premultiplied);
+    red.fill(QColor(255, 0, 0, 128));
+    auto blueBuffer = m_shm->createBuffer(blue);
+    s->attachBuffer(blueBuffer.data());
+    s->setScale(2);
+    s->commit(Surface::CommitFlag::None);
+    QVERIFY(sizeChangedSpy.wait());
+    QCOMPARE(sizeChangedSpy.count(), 1);
+    QCOMPARE(scaleChangedSpy.count(), 1);
+    QCOMPARE(serverSurface->scale(), 2);
+    QCOMPARE(serverSurface->size(), QSize(25,25));
 }
 
 void TestWaylandSurface::testDestroy()
@@ -858,7 +914,7 @@ void TestWaylandSurface::testDestroyAttachedBuffer()
     // let's damage this surface
     QSignalSpy damagedSpy(serverSurface, &SurfaceInterface::damaged);
     QVERIFY(damagedSpy.isValid());
-    QImage image(QSize(100, 100), QImage::Format_ARGB32);
+    QImage image(QSize(100, 100), QImage::Format_ARGB32_Premultiplied);
     image.fill(Qt::red);
     s->attachBuffer(m_shm->createBuffer(image));
     s->damage(QRect(0, 0, 100, 100));
@@ -899,7 +955,7 @@ void TestWaylandSurface::testDestroyWithPendingCallback()
     QVERIFY(serverSurface);
 
     // now render to it
-    QImage img(QSize(10, 10), QImage::Format_ARGB32);
+    QImage img(QSize(10, 10), QImage::Format_ARGB32_Premultiplied);
     img.fill(Qt::black);
     auto b = m_shm->createBuffer(img);
     s->attachBuffer(b);
@@ -956,5 +1012,84 @@ void TestWaylandSurface::testDisconnect()
     m_queue->destroy();
 }
 
+void TestWaylandSurface::testOutput()
+{
+    // This test verifies that the enter/leave are sent correctly to the Client
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+    qRegisterMetaType<KWayland::Client::Output*>();
+    QScopedPointer<Surface> s(m_compositor->createSurface());
+    QVERIFY(!s.isNull());
+    QVERIFY(s->isValid());
+    QVERIFY(s->outputs().isEmpty());
+    QSignalSpy enteredSpy(s.data(), &Surface::outputEntered);
+    QVERIFY(enteredSpy.isValid());
+    QSignalSpy leftSpy(s.data(), &Surface::outputLeft);
+    QVERIFY(leftSpy.isValid());
+    // wait for the surface on the Server side
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    QVERIFY(surfaceCreatedSpy.wait());
+    auto serverSurface = surfaceCreatedSpy.first().first().value<SurfaceInterface*>();
+    QVERIFY(serverSurface);
+    QCOMPARE(serverSurface->outputs(), QVector<OutputInterface*>());
+
+    // create another registry to get notified about added outputs
+    Registry registry;
+    registry.setEventQueue(m_queue);
+    QSignalSpy allAnnounced(&registry, &Registry::interfacesAnnounced);
+    QVERIFY(allAnnounced.isValid());
+    registry.create(m_connection);
+    QVERIFY(registry.isValid());
+    registry.setup();
+    QVERIFY(allAnnounced.wait());
+    QSignalSpy outputAnnouncedSpy(&registry, &Registry::outputAnnounced);
+    QVERIFY(outputAnnouncedSpy.isValid());
+
+    auto serverOutput = m_display->createOutput(m_display);
+    serverOutput->create();
+    QVERIFY(outputAnnouncedSpy.wait());
+    QScopedPointer<Output> clientOutput(registry.createOutput(outputAnnouncedSpy.first().first().value<quint32>(), outputAnnouncedSpy.first().last().value<quint32>()));
+    QVERIFY(clientOutput->isValid());
+    m_connection->flush();
+    m_display->dispatchEvents();
+
+    // now enter it
+    serverSurface->setOutputs(QVector<OutputInterface*>{serverOutput});
+    QCOMPARE(serverSurface->outputs(), QVector<OutputInterface*>{serverOutput});
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.count(), 1);
+    QCOMPARE(enteredSpy.first().first().value<Output*>(), clientOutput.data());
+    QCOMPARE(s->outputs(), QVector<Output*>{clientOutput.data()});
+
+    // adding to same should not trigger
+    serverSurface->setOutputs(QVector<OutputInterface*>{serverOutput});
+
+    // leave again
+    serverSurface->setOutputs(QVector<OutputInterface*>());
+    QCOMPARE(serverSurface->outputs(), QVector<OutputInterface*>());
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(enteredSpy.count(), 1);
+    QCOMPARE(leftSpy.count(), 1);
+    QCOMPARE(leftSpy.first().first().value<Output*>(), clientOutput.data());
+    QCOMPARE(s->outputs(), QVector<Output*>());
+
+    // leave again should not trigger
+    serverSurface->setOutputs(QVector<OutputInterface*>());
+
+    // and enter again, just to verify
+    serverSurface->setOutputs(QVector<OutputInterface*>{serverOutput});
+    QCOMPARE(serverSurface->outputs(), QVector<OutputInterface*>{serverOutput});
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 1);
+
+    //delete output client is on.
+    //client should get an exit and be left on no outputs (which is allowed)
+    serverOutput->deleteLater();
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(serverSurface->outputs(), QVector<OutputInterface*>());
+}
+
 QTEST_GUILESS_MAIN(TestWaylandSurface)
 #include "test_wayland_surface.moc"
