@@ -25,6 +25,8 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include <QDebug>
 #include <QDir>
 #include <QFileSystemWatcher>
+#include <QMutex>
+#include <QMutexLocker>
 #include <QSocketNotifier>
 #include <qpa/qplatformnativeinterface.h>
 // Wayland
@@ -36,7 +38,7 @@ namespace KWayland
 namespace Client
 {
 
-class ConnectionThread::Private
+class Q_DECL_HIDDEN ConnectionThread::Private
 {
 public:
     Private(ConnectionThread *q);
@@ -53,10 +55,18 @@ public:
     QScopedPointer<QFileSystemWatcher> socketWatcher;
     bool serverDied = false;
     bool foreign = false;
+    QMetaObject::Connection eventDispatcherConnection;
+    int error = 0;
+    static QVector<ConnectionThread*> connections;
+    static QMutex mutex;
 private:
     ConnectionThread *q;
 };
 
+QVector<ConnectionThread*> ConnectionThread::Private::connections = QVector<ConnectionThread*>{};
+QMutex ConnectionThread::Private::mutex{QMutex::Recursive};
+
+
 ConnectionThread::Private::Private(ConnectionThread *q)
     : socketName(QString::fromUtf8(qgetenv("WAYLAND_DISPLAY")))
     , runtimeDir(QString::fromUtf8(qgetenv("XDG_RUNTIME_DIR")))
@@ -65,10 +75,18 @@ ConnectionThread::Private::Private(ConnectionThread *q)
     if (socketName.isEmpty()) {
         socketName = QStringLiteral("wayland-0");
     }
+    {
+        QMutexLocker lock(&mutex);
+        connections << q;
+    }
 }
 
 ConnectionThread::Private::~Private()
 {
+    {
+        QMutexLocker lock(&mutex);
+        connections.removeOne(q);
+    }
     if (display && !foreign) {
         wl_display_flush(display);
         wl_display_disconnect(display);
@@ -108,7 +126,17 @@ void ConnectionThread::Private::setupSocketNotifier()
             if (!display) {
                 return;
             }
-            wl_display_dispatch(display);
+            if (wl_display_dispatch(display) == -1) {
+                error = wl_display_get_error(display);
+                if (error != 0) {
+                    if (display) {
+                        free(display);
+                        display = nullptr;
+                    }
+                    emit q->errorOccurred();
+                    return;
+                }
+            }
             emit q->eventsRead();
         }
     );
@@ -146,6 +174,7 @@ void ConnectionThread::Private::setupSocketFileWatcher()
                         qCDebug(KWAYLAND_CLIENT) << "Socket reappeared";
                         socketWatcher.reset();
                         serverDied = false;
+                        error = 0;
                         q->initConnection();
                     }
                 }
@@ -159,7 +188,7 @@ ConnectionThread::ConnectionThread(QObject *parent)
     : QObject(parent)
     , d(new Private(this))
 {
-    connect(QCoreApplication::eventDispatcher(), &QAbstractEventDispatcher::aboutToBlock, this,
+    d->eventDispatcherConnection = connect(QCoreApplication::eventDispatcher(), &QAbstractEventDispatcher::aboutToBlock, this,
         [this] {
             if (d->display) {
                 wl_display_flush(d->display);
@@ -168,7 +197,10 @@ ConnectionThread::ConnectionThread(QObject *parent)
         Qt::DirectConnection);
 }
 
-ConnectionThread::~ConnectionThread() = default;
+ConnectionThread::~ConnectionThread()
+{
+    disconnect(d->eventDispatcherConnection);
+}
 
 ConnectionThread *ConnectionThread::fromApplication(QObject *parent)
 {
@@ -186,6 +218,7 @@ ConnectionThread *ConnectionThread::fromApplication(QObject *parent)
     ConnectionThread *ct = new ConnectionThread(parent);
     ct->d->foreign = true;
     ct->d->display = display;
+    connect(native, &QObject::destroyed, ct, &ConnectionThread::connectionDied);
     return ct;
 }
 
@@ -254,5 +287,20 @@ void ConnectionThread::roundtrip()
     wl_display_roundtrip(d->display);
 }
 
+bool ConnectionThread::hasError() const
+{
+    return d->error != 0;
+}
+
+int ConnectionThread::errorCode() const
+{
+    return d->error;
+}
+
+QVector<ConnectionThread*> ConnectionThread::connections()
+{
+    return Private::connections;
+}
+
 }
 }
